from typing import Optional

import trio

from ..utils import url_version
from .installer import make_installer
from .manual_version import ManualVersion

INSTALLER_LOCK = trio.Lock()


class SystemPackage(ManualVersion):
    system: str
    url: Optional[str] = None
    auto_updates: bool = False

    async def installer(self):
        return await make_installer(self.driver)

    def derive_name(self) -> str:
        return self.system

    async def get_remote_version(self) -> str:
        if self.url:
            return await url_version(self.url)
        if self.auto_updates:
            return "latest"
        return await (await self.installer()).latest_version(self.system)

    async def local_version(self) -> Optional[str]:
        if self.url:
            return self.cached_version
        installer = await self.installer()
        version = await installer.installed_version(self.system)
        if self.auto_updates:
            return "latest" if version else None
        return version

    async def install(self) -> None:
        async with INSTALLER_LOCK:
            installer = await self.installer()
            if self.url:
                await installer.install(self.url)
                await self.cache_version()
            elif await self.local_version():
                await installer.upgrade(self.system)
            else:
                await installer.install(self.system)

        await super().install()



from collections.abc import Sequence
from typing import Any, Union

from pydantic import TypeAdapter

from ..driver import Driver
from ..state import DB
from .base import Package, PackageArgs
from .system import SystemPackage

AnyPackageT = SystemPackage

AnyPackage: TypeAdapter[AnyPackageT] = TypeAdapter(AnyPackageT)

AnyPackageList = TypeAdapter(list[AnyPackageT])


def parse_package(package: Any, *, db: DB, driver: Driver) -> Package:
    package = TypeAdapter(dict).validate_python(package)

    return AnyPackage.validate_python({**package, "db": db, "driver": driver})


def parse_packages(packages: Any, *, db: DB, driver: Driver) -> Sequence[Package]:
    packages = TypeAdapter(list[dict]).validate_python(packages)

    return AnyPackageList.validate_python(
        [{**package, "db": db, "driver": driver} for package in packages]
    )



from typing import Optional

from .base import Installer


class Apt(Installer):
    async def install(self, package: str) -> None:
        await self.driver.run(
            "sudo",
            "env",
            "DEBIAN_FRONTEND=noninteractive",
            "apt",
            "install",
            "--yes",
            package,
        )

    async def upgrade(self, package: str) -> None:
        await self.install(package)

    async def latest_version(self, package: str) -> str:
        result = await self.driver.run_output_(
            "apt-cache",
            "show",
            "--quiet",
            "--no-all-versions",
            package,
        )
        if not result.ok:
            raise ValueError(f"Cannot determine version for: {package}.")
        for line in result.output.splitlines():
            line = line.strip()
            if line.startswith("Version:"):
                return line.split(": ", 1)[-1]
        raise Exception(f"Cannot parse apt output: {result.output}")  # pragma: no cover

    async def installed_version(self, package: str) -> Optional[str]:
        return (
            await self.driver.run_output_(
                "dpkg-query", "--showformat", "${Version}", "--show", package
            )
        ).output



import subprocess
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable, Sequence
from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Optional, cast

from trio import run_process

from .utils import RunArg, T, async_cached


class OS(ABC):
    @abstractmethod
    def switch_(self, *, linux: Callable[["Linux"], T], macos: Callable[[], T]) -> T:
        pass

    def switch(self, *, linux: T, macos: T) -> T:
        return self.switch_(linux=lambda _: linux, macos=lambda: macos)


@dataclass
class Linux(OS):
    distribution: str

    def switch_(self, *, linux: Callable[["Linux"], T], macos: Callable[[], T]) -> T:
        return linux(self)

    RELEASE_FILE = "/etc/os-release"

    @classmethod
    async def get_distribution(cls, driver: "Driver") -> str:
        release_file = await driver.run_output("cat", Path(cls.RELEASE_FILE))
        for line in release_file.splitlines():
            k, v = line.split("=", 1)
            if k == "ID":
                if v.startswith('"') and v.endswith('"'):
                    v = v[1:-1]
                return v

        raise ValueError(f"Cannot find distribution ID in {cls.RELEASE_FILE}.")


@dataclass
class MacOS(OS):
    def switch_(self, *, linux: Callable[["Linux"], T], macos: Callable[[], T]) -> T:
        return macos()


Architecture = Literal["x86_64", "aarch64"]


@dataclass
class RunResult:
    ok: bool


@dataclass
class RunResultOutput(RunResult):
    ok: bool
    output: str


class Driver(ABC):
    @abstractmethod
    async def run_(
        self,
        *args: RunArg,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        pass

    async def run(
        self,
        *args: RunArg,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        silent: bool = False,
    ) -> None:
        await self.run_(*args, input=input, silent=silent)

    async def run_ok(self, *args: RunArg) -> bool:
        result = await self.run_(*args, check=False)
        return result.ok

    async def run_output(self, *args: RunArg, silent: bool = False) -> str:
        result = await self.run_(*args, capture_output=True, silent=silent)
        return cast(RunResultOutput, result).output

    async def run_output_(self, *args: RunArg) -> RunResultOutput:
        result = await self.run_(*args, capture_output=True, check=False, silent=True)
        return cast(RunResultOutput, result)

    async def executable_exists(self, executable: str) -> bool:
        return await self.run_ok("sh", "-c", f"command -v {executable}")

    async def find_executable(self, *executables: str) -> str:
        for candidate in executables:
            if await self.executable_exists(candidate):
                return candidate
        raise Exception(f"None of {', '.join(executables)} found in PATH.")

    async def is_file(self, path: Path) -> bool:
        return await self.run_ok("test", "-f", path)

    async def is_executable(self, path: Path) -> bool:
        return await self.run_ok("test", "-x", path)

    async def is_dir(self, path: Path) -> bool:
        return await self.run_ok("test", "-d", path)

    async def username(self) -> str:
        return await self.run_output("whoami")

    @async_cached
    async def home(self) -> Path:
        result = await self.run_output("sh", "-c", "eval echo ~")
        return Path(result)

    @asynccontextmanager
    async def tempfile(
        self, kind: Optional[Literal["directory"]] = None
    ) -> AsyncIterator[Path]:
        args = []
        if kind == "directory":
            args.append("-d")

        path = Path(await self.run_output("mktemp", *args))
        try:
            yield path
        finally:
            await self.rm(path)

    async def makedirs(self, path: Path) -> None:
        await self.run("mkdir", "-p", path)

    async def copy(self, source: Path, target: Path) -> None:
        await self.makedirs(target.parent)
        await self.rm(target)
        await self.run("cp", "-R", "-f", source, target)

    async def rm(self, path: Path) -> None:
        await self.run("rm", "-r", "-f", path)

    async def make_executable(self, path: Path) -> None:
        await self.run("chmod", "+x", path)

    async def write_file(self, path: Path, content: str) -> None:
        await self.makedirs(path.parent)
        await self.rm(path)
        await self.run("cp", "/dev/stdin", path, input=content.encode())

    async def link(
        self,
        source: Path,
        target: Path,
    ) -> None:
        await self.makedirs(target.parent)
        await self.rm(target)
        await self.run("ln", "-s", "-f", source, target)

    @async_cached
    async def os(self) -> OS:
        os_type = await self.run_output("uname")
        if os_type == "Linux":
            distribution = await Linux.get_distribution(self)
            return Linux(distribution=distribution)
        elif os_type == "Darwin":
            return MacOS()
        else:
            raise ValueError(f"Unsupported OS type {os_type}.")  # pragma: no cover

    @async_cached
    async def architecture(self) -> Architecture:
        result = await self.run_output("uname", "-m")
        if result == "x86_64":
            return "x86_64"
        elif result in {"arm64", "aarch64"}:
            return "aarch64"
        else:
            raise ValueError(f"Unsupported architecture {result}.")  # pragma: no cover


class LocalDriver(Driver):
    def prepare_command(self, args: Sequence[RunArg]) -> list[RunArg]:
        return list(args)

    async def run_(
        self,
        *args: RunArg,
        check: bool = True,
        input: Optional[bytes] = None,  # pylint:disable=redefined-builtin
        capture_output: bool = False,
        silent: bool = False,
    ) -> RunResult:
        command = self.prepare_command(args)

        if not check or silent:
            stderr = subprocess.DEVNULL
        else:
            stderr = None

        result = await run_process(
            command,
            check=check,
            stdin=input,
            capture_stdout=True,
            stderr=stderr,
            **self.run_args(),
        )

        ok = result.returncode == 0
        if capture_output:
            output = result.stdout.decode().strip()
            return RunResultOutput(ok=ok, output=output)
        else:
            return RunResult(ok=ok)

    def run_args(self) -> dict[str, Any]:
        return {}



from collections.abc import AsyncIterable, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml

from .config import MatchConfig, parse_config
from .driver import Driver
from .package import Package, parse_packages
from .parallel import PartialException, PartialResults
from .state import DB, VERSIONS


@dataclass
class InstallResult:
    installed: list[Package]
    failed: list[tuple[Package, BaseException]]

    def raise_for_failures(self) -> None:  # pragma: no cover
        if self.failed:
            for _, exc in self.failed:
                raise exc


@dataclass
class Manager:
    db: DB
    driver: Driver
    data_path: Path

    def load_data(self, *path: str) -> Any:
        with open(self.data_path / Path(*path)) as f:
            return yaml.safe_load(f)

    def load_config(self) -> list[MatchConfig]:
        return parse_config(self.load_data("mybox.yaml"), driver=self.driver)

    def load_component(self, component: str) -> Sequence[Package]:
        packages = self.load_data("packages", f"{component}.yaml")
        return parse_packages(packages, db=self.db, driver=self.driver)

    def load_components(self, components: frozenset[str]) -> list[Package]:
        return [
            pkg for component in components for pkg in self.load_component(component)
        ]

    async def cleanup(self, packages: list[Package]) -> None:
        package_names = set(package.name for package in packages)

        versions = VERSIONS(self.db)

        all_versions = list(versions.find_ids())

        for package, _ in all_versions:
            if package not in package_names:
                versions.delete(id=package)

    async def install(self) -> InstallResult:
        config = self.load_config()
        components = await MatchConfig.components(config)
        packages = self.load_components(frozenset(components))

        return await self.install_packages(packages)

    async def install_packages(self, packages: list[Package]) -> InstallResult:
        try:
            results = []
            for pkg in packages:
                async for package in self.install_package(pkg):
                    results.append(package)

            await self.cleanup(packages)

            return InstallResult(installed=results, failed=[])
        except PartialResults as e:
            installed: list[Package] = []
            failed: list[tuple[Package, BaseException]] = []

            for package, result in zip(packages, e.results):
                if isinstance(result, PartialException):
                    failed.append((package, result.exception))
                else:
                    installed.extend(result.result)

            return InstallResult(installed=installed, failed=failed)

    async def install_package(self, package: Package) -> AsyncIterable[Package]:
        if not await package.applicable():
            return

        if await package.is_installed():
            return

        await package.install()

        yield package



from dataclasses import dataclass
from typing import Generic

from .utils import T


@dataclass
class PartialSuccess(Generic[T]):
    result: T


@dataclass
class PartialException:
    exception: BaseException


PartialResult = PartialException | PartialSuccess[T]


@dataclass
class PartialResults(Exception, Generic[T]):
    results: list[PartialResult[T]]
